def accuracy_topk(logits, y, k):
    acc = (logits.topk(max((1, k)), 1, True, True)[1] == y.view(-1, 1)).sum().float().item()/logits.shape[0]
    return acc


def accuracy(logits, y):
    return accuracy_topk(logits, y, 1)


def accuracy_top3(logits, y):
    return accuracy_topk(logits, y, 3)



from torch.utils.data import DataLoader
from .dry import Dry


def DataLoaders(root, dataset_class, transform_train, transform_test=None, batch_size=32, num_workers=2, pin_memory=False, compile=True, logger=None):
    if transform_test is None:
        transform_test = transform_train

    trainset = Dry.Dataset(dataset_class)  # create dry dataset parameters.
    trainset.root = root
    trainset.train = True
    trainset.transform = transform_train
    testset = Dry.Dataset(dataset_class)  # create dry dataset parameters.
    testset.root = root
    testset.train = False
    testset.transform = transform_test
    valset = Dry.Dataset(dataset_class)
    valset.root = root
    valset.train = False
    valset.transform = transform_test

    trainloader = Dry.Args(DataLoader)  # init trainloader
    trainloader.dataset = trainset
    trainloader.batch_size = batch_size
    trainloader.shuffle = True
    trainloader.num_workers = num_workers
    trainloader.pin_memory = pin_memory

    valloader = Dry.Args(DataLoader)  # init valloader
    valloader.dataset = valset
    valloader.batch_size = batch_size
    valloader.shuffle = True
    valloader.num_workers = num_workers
    valloader.pin_memory = pin_memory

    testloader = Dry.Args(DataLoader)  # init testloader
    testloader.dataset = testset
    testloader.batch_size = batch_size
    testloader.shuffle = False
    testloader.num_workers = num_workers
    testloader.pin_memory = pin_memory

    if compile:
        trainloader = trainloader.compile(logger=logger)
        valloader = valloader.compile(logger=logger)
        testloader = testloader.compile(logger=logger)

    return trainloader, valloader, testloader



import torch

from . import functional


class TrainLooper():
    def __init__(self, trainloader, model, criterion=None, optimizer=None,
                 max_epoch=None, start_epoch=0, cuda=True, check_type_tensor=False):
        self.trainloader = trainloader
        self.model = model
        if criterion is None:
            criterion = torch.nn.CrossEntropyLoss()
        self.criterion = criterion
        if optimizer is None:
            optimizer = torch.optim.SGD(model.parameters(), lr=1e-2)
        self.optimizer = optimizer
        self.cuda = cuda
        self.check_type_tensor = check_type_tensor

        self.max_epoch = max_epoch
        self.start_epoch = start_epoch
        self._looper = None

    def _main_loop(self):
        epoch = self.start_epoch
        while True:
            self.epoch = epoch
            for iteration, batch in enumerate(self.trainloader):  # batch loop
                loss = functional.train_step(  # one step forward & backward
                    batch=batch,
                    model=self.model,
                    criterion=self.criterion,
                    optimizer=self.optimizer,
                    cuda=self.cuda,
                    check_type_tensor=self.check_type_tensor
                )
                self.iteration = iteration
                yield loss
            epoch += 1
            if self.max_epoch is not None:
                if epoch > self.max_epoch:
                    return

    @property
    def epoch_finished(self):
        if self.iteration + 1 == len(self.trainloader):
            return True
        return False

    def steps(self, **kwargs):
        while True:
            yield(self.step(**kwargs))

    def step(self, iteration=0, epoch=0):
        if self._looper is None:
            self._looper = self._main_loop()
            self.epoch = self.start_epoch
        if 0 < epoch < 1:
            iteration = int(len(self.trainloader)*epoch)  # convert float epoch to iterations
            epoch = 0
        if iteration + epoch == 0:
            iteration = 1  # step for at least 1 iteration.

        target_epoch = self.epoch + epoch
        loss = 0.0
        total = 0
        while target_epoch > self.epoch:
            loss += next(self._looper)
            total += 1
        for _ in range(iteration):
            loss += next(self._looper)
            total += 1
            if self.epoch_finished:
                break
        return loss/total

    def __iter__(self):
        yield self.step(epoch=1)


class ValLooper():
    def __init__(self, valloader, model, criterion=None, cuda=True, check_type_tensor=False):
        self.valloader = valloader
        self.model = model
        if criterion is None:
            criterion = torch.nn.CrossEntropyLoss()
        self.criterion = criterion
        self._looper = None
        self.cuda = cuda
        self.check_type_tensor = check_type_tensor

    def _main_loop(self):
        while True:
            for batch in self.valloader:  # batch loop
                loss = functional.validation_step(  # one step forward
                    batch=batch,
                    model=self.model,
                    criterion=self.criterion,
                    cuda=self.cuda,
                    check_type_tensor=self.check_type_tensor
                )
                yield loss

    def step(self, iteration=4, epoch=0):
        if self._looper is None:
            self._looper = self._main_loop()

        if epoch > 0:
            iteration = int(len(self.trainloader)*epoch)  # convert epoch to iterations

        loss = 0.0
        for _ in range(iteration):
            loss += next(self._looper)
        return loss/iteration



import torch


def _check_batch_data(batch, cuda, check_type_tensor):
    x, target = batch
    if check_type_tensor:
        if not isinstance(x, torch.Tensor):
            x = torch.tensor(x)
        if not isinstance(target, torch.Tensor):
            target = torch.tensor(target)
    if cuda:
        x = x.cuda()
        target = target.cuda()
    return x, target


def train_step(batch, model: torch.nn.Module, criterion, optimizer: torch.optim.Optimizer, cuda=True, check_type_tensor=True):
    x, target = _check_batch_data(batch, cuda=cuda, check_type_tensor=check_type_tensor)  # check inputs
    if not model.training:  # switch model mode
        model.train()
    optimizer.zero_grad()
    y = model(x)
    loss = criterion(y, target)
    loss.backward()
    optimizer.step()
    return loss.item()


def validation_step(batch, model: torch.nn.Module, criterion=None, cuda=True, check_type_tensor=True):
    x, target = _check_batch_data(batch, cuda=cuda, check_type_tensor=check_type_tensor)  # check inputs
    _ret = None
    with torch.no_grad():  # forward without gradient
        if model.training:  # switch model mode
            model.eval()
        y = model(x)
        if criterion:  # return the loss if a criterion function is specified.
            loss = criterion(y, target)
            _ret = loss.item()
        else:
            _ret = y.data
    return _ret



import time


class Time():
    def __init__(self, format="[%Y-%m-%d %H:%M:%S]"):
        self.format = format

    def asctime(self):
        return time.asctime(time.localtime())

    def now(self):
        return time.strftime(self.format, time.localtime())



import torch
from .step import _check_batch_data


def _correct_topk(inputs: torch.Tensor, target: torch.Tensor, k: int):
    if k > 1:
        inputs = inputs.topk(k, 1, True, True)[1]
        target = target.view(-1, 1)
    else:
        _, inputs = torch.max(inputs, 1)
    return inputs.eq(target).sum().float().item()


def accuracy(testloader: torch.utils.data.DataLoader, model: torch.nn.Module, topk=1, cuda=True, check_type_tensor=True):
    with torch.no_grad():
        if model.training:
            model.eval()
        correct = 0.0
        total = 0.0
        for batch in testloader:
            x, target = _check_batch_data(batch, cuda=cuda, check_type_tensor=check_type_tensor)
            outputs = model(x)
            total += target.size(0)
            correct += _correct_topk(outputs, target, topk)
    acc = correct / total
    return acc



from . import auto
from . import datasets
from . import functional
from . import transforms
from . import dock
from .dock import visualize

__all__ = ['auto', 'datasets', 'functional', 'transforms', 'dock', 'visualize']





class Logger():
    def __init__(self, fp=None):
        self.fp = fp

    def __call__(self, msg):
        print(msg)
        if self.fp is not None:
            with open(self.fp, 'a') as f:
                f.write(f'{msg}\n')




import os
import re
import time

import paramiko

from .local import pack


class SSHConnector():
    def __init__(self, ip, user, password, port=22):
        self.ssh = None
        self.ip = ip
        self.port = port
        self.user = user
        self.password = password
        self.cmd_location = '~'

    def __str__(self):
        return f'{self.user}@{self.ip}:{self.port}'

    def log(self, msg, **kwargs):
        print(f'\t[{self.ip}] {msg}', **kwargs)

    def __enter__(self):
        if self.ssh is None:
            self.connect()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def connect(self):
        ssh = paramiko.SSHClient()  # 创建sshclient
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())  # 指定当对方主机没有本机公钥的情况时应该怎么办，AutoAddPolicy表示自动在对方主机保存下本机的秘钥
        ssh.connect(self.ip, self.port, self.user, self.password)
        self.log(f'connected to {self.__str__()}')
        self.ssh = ssh
        return self

    def close(self):
        self.log('ssh closed.')
        self.ssh.close()
        self.ssh = None

    def cd(self, location):
        if self.ssh is None:
            self.connect()
        self.log(f'server:{self.cmd_location}$ cd {location}')
        _, stdout, stderr = self.ssh.exec_command(f'cd {location};pwd',)
        self.cmd_location = stdout.readlines()[0].replace('\n', '')
        for line in stderr:
            self.log(f'{self.cmd_location}$ {line}', end='')
        return stdout, stderr

    def __call__(self, command):
        return self.exec_command(command)

    def exec_command(self, command):
        if self.ssh is None:
            self.connect()
        if command.startswith('cd'):
            return self.cd(command[3:])
        self.log(f'server:{self.cmd_location}$ {command}')
        _, stdout, stderr = self.ssh.exec_command(f'. /etc/profile;. .bashrc;cd {self.cmd_location};{command}')
        out = stdout.readlines()
        for line in out:
            self.log(f'server:{self.cmd_location}$ {line}', end='')
        err = stderr.readlines()
        for line in err:
            self.log(f'server:{self.cmd_location}$ {line}', end='')
        return out, err

    @staticmethod
    def sftp_callback(transferred, total):
        print(f'\t[sftp] transferred/total: {transferred}/{total} ({transferred/total:%})')

    def upload(self, local_file, remote_file):
        if self.ssh is None:
            self.connect()
        _dir = re.sub(remote_file.split('/')[-1], '', remote_file)
        self.exec_command(f'mkdir {_dir}')
        sftp = self.ssh.open_sftp()
        sftp.put(local_file, remote_file, callback=self.sftp_callback)


class Server():
    def __init__(self, remote_root, **kwargs):
        self.ssh = SSHConnector(**kwargs)
        self.cache_dir = 'build'
        self.remote_root = remote_root
        self.fsize_limit = 1024

    def deploy(self, src, run='main.py', conda_env=None):
        deploy_pack = pack(src, self.cache_dir)
        fsize = os.path.getsize(deploy_pack)
        f_mb = fsize/float(1024)/float(1024)
        print(f'package size: {f_mb*1024:.6f}KB ({f_mb:.4f}MB)')
        if f_mb > self.fsize_limit:
            print('src too large for ssh.')
            raise NotImplementedError
        print('transfering deploy package to remote ssh ..')
        self.ssh.log('cleanning target dir..')
        self.ssh.exec_command(f'rm -r {self.remote_root}')
        self.ssh.log('sftp connection established, uploading ..')
        self.ssh.upload(deploy_pack, f'{self.remote_root}/deploy_pack.tar.gz')
        self.ssh.exec_command(f'cd {self.remote_root}')
        self.ssh.exec_command(f'tar -zxvf deploy_pack.tar.gz')
        if conda_env:
            self.ssh.exec_command(f'source activate {conda_env};nohup python {run} >log.d 2>&1 &')
        else:
            self.ssh.exec_command(f'screen -s labvision-remote;nohup python {run} >log.d 2>&1 &')
        self.ssh.log(f'successfully deployed, running in background now.')
        self.ssh.log(f'tracking python process:')
        self.ssh.exec_command(f'ps -u {self.ssh.user}|grep python')
        self.ssh.log(f"you can use 'cd {self.remote_root};cat log.d' to see the log.")
        print(f'removing cached deploy package: {deploy_pack}')
        os.remove(deploy_pack)

    def exec_command(self, command):
        return self.ssh(command)

    def __call__(self, command):
        return self.ssh(command)

    def close(self, latency=1):
        self.ssh.log(f'ssh close in {latency} secs ..')
        time.sleep(latency)
        self.ssh.close()

    def __enter__(self):
        self.ssh.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()





import re
import inspect
import os
import random
import string
import time
import json

import torch
from labvision import functional
from torch.autograd import Variable
from torch.utils.data import DataLoader

from .utils import manual_seed


class MissingArgsException(Exception):
    def __init__(self, key, msg):
        super().__init__()
        msg = f'Missing key \'{key}\' in config\n({msg})'

    def __str__(self):
        return f'{self.msg}'


class Dry():
    class Args():
        def __init__(self, compile_class):
            parameters = inspect.signature(compile_class).parameters
            parameter_dict = {k: self.__defaultargs__(v) for k, v in parameters.items()}
            self.compile_class = compile_class
            self.__dict__.update(parameter_dict)

        def update(self, **kwargs):
            self.__dict__.update(kwargs)

        @staticmethod
        def __defaultargs__(v):
            if v.default is not inspect._empty:
                return v.default
            if v.kind == inspect.Parameter.VAR_KEYWORD:
                return {}
            if v.kind == inspect.Parameter.VAR_POSITIONAL:
                return []

        def __str__(self):
            return str(self.__dict__)

        def log(self, msg, logger=None):
            if logger is None:
                print(msg)
            else:
                logger.log(msg)

        def compile(self, logger=None, **kwargs):
            args = {k: v for k, v in self.__dict__.items() if k != 'compile_class' and k != 'kwargs'}
            args.update(kwargs)
            try:
                self.log(f'==> Compiling: {self.compile_class}', logger=logger)
                for k, v in args.items():
                    if isinstance(v, Dry.Args):
                        v = v.compile(logger=logger)
                    self.log(f'  -> Running args hook: {k} = {v}', logger=logger)
                    assert v is not inspect._empty
            except Exception:
                raise MissingArgsException(k, 'This value must be specified.')
            return self.compile_class(**args)

    class Optimizer(Args):
        def __init__(self, compile_class):
            super().__init__(compile_class)
            self.lr = 1e-3
            self.weight_decay = 5e-4

    class Dataset(Args):
        def __init__(self, compile_class):
            super().__init__(compile_class)
            self.root = None
            self.transform = None


class Core():
    def __init__(self, root='.'):
        self.__model = None
        self.__optimizer = None
        self.__trainset = None
        self.__testset = None
        self.__valset = None
        self.seed = None
        self.root = root
        self.batch_size = 16
        self.num_workers = 2
        self.netdata_log_path = None
        self.slave_log_path = None
        self.auto_log = True

    @property
    def model(self): return self.__model
    @model.setter
    def model(self, x):
        if inspect.isclass(x):
            x = Dry.Args(x)
        self.__model = x

    @property
    def optimizer(self): return self.__optimizer
    @optimizer.setter
    def optimizer(self, x):
        if inspect.isclass(x):
            x = Dry.Optimizer(x)
        self.__optimizer = x

    @property
    def trainset(self): return self.__trainset
    @trainset.setter
    def trainset(self, x):
        if inspect.isclass(x):
            x = Dry.Dataset(x)
        self.__trainset = x

    @property
    def testset(self): return self.__testset
    @testset.setter
    def testset(self, x):
        if inspect.isclass(x):
            x = Dry.Dataset(x)
        self.__testset = x

    @property
    def valset(self): return self.__valset
    @valset.setter
    def valset(self, x):
        if inspect.isclass(x):
            x = Dry.Dataset(x)
        self.__valset = x

    def set_datasets(self, dataset_class, **kwargs):
        self.trainset = Dry.Dataset(dataset_class)
        self.trainset.update(**kwargs)
        self.testset = Dry.Dataset(dataset_class)
        self.testset.update(**kwargs)

    def push(self, name, value, dry=False, **kwargs):
        if dry and inspect.isclass(value):
            value = Dry.Args(value)
            value.update(**kwargs)
        self.__dict__[name] = value

    def compile(self):
        return Slave(self)

    def freeze(self, fp):
        _dir = re.sub(fp.split('/')[-1], '', fp)
        if not os.path.exists(_dir):
            os.makedirs(_dir)
        torch.save(self, fp)
        return fp


class Slave():
    class CurrentStatus():
        def __init__(self, frozen_dict=None):
            if frozen_dict:
                self.state_dict = frozen_dict
            else:
                self.state_dict = dict(
                    hash=self.__genhash__(),
                    train={},
                    val={},
                )

        @property
        def train(self):
            return self.state_dict['train']

        @property
        def val(self):
            return self.state_dict['val']

        @staticmethod
        def __genhash__():
            """
                generates a hash with ASCII letters and digits,
                always starts with a letter (for markdown usage).
            """
            random.seed()
            _hash_head = ''.join(random.sample(string.ascii_letters, 1))
            _hash_body = ''.join(random.sample(string.ascii_letters+string.digits, 7))
            return _hash_head+_hash_body

        def epoch_finished(self):
            return self.train['epoch_finished']

        @property
        def hash(self):
            return self.state_dict['hash']

        @property
        def iter(self):
            return self.train['iter']

        @property
        def epoch(self):
            if 'epoch' not in self.train:
                return 0
            return self.train['epoch']

    def __init__(self, core, cuda=True):
        self.core = core
        self.cuda = cuda
        self.status = self.CurrentStatus()
        self.step_iters = 200
        self.val_iters = 16
        self._eval_hooks = []
        self.__train_generator = None
        self.__val_generator = None
        self.under_test = False
        self.netdata_state = dict(hash=self.hash)
        self.compile(core)

    def compile(self, core: Core):
        self.root = core.root
        if not os.path.exists(core.root):
            os.makedirs(core.root)
            self.log(f'path not exists, creating dir: {core.root}')
        self.log('start compiling ..')
        assert isinstance(core.model, torch.nn.Module)
        assert core.trainset is not None
        assert core.testset is not None
        if core.seed is None:
            random.seed()
            self.seed = random.randint(0, 1e8)
            self.log(f'initialize seed = {self.seed}')
        else:
            self.seed = core.seed
        self.init_seed(self.seed)
        if isinstance(core.model, Dry.Args):
            self.log(f'compiling model from {core.model.compile_class} ..')
            self.model = core.model.compile(logger=self)
        else:
            self.model = core.model
        if self.cuda:
            self.log(f'transfer model to cuda device ..')
            self.model = self.model.cuda()
        self.optimizer = core.optimizer
        if self.optimizer is None:
            self.optimizer = Dry.Optimizer(torch.optim.SGD)
            self.log(f'using default optimizer as <torch.optim.SGD>')
        if isinstance(self.optimizer, Dry.Optimizer):
            self.log(f'compiling optimizer from {self.optimizer.compile_class} ..')
            self.optimizer = self.optimizer.compile(params=self.model.parameters(), logger=self)
        self.log(f'compiling datasets ..')
        trainset = core.trainset
        testset = core.testset
        valset = core.valset
        if valset is None:
            valset = testset
        if isinstance(trainset, Dry.Dataset):
            trainset = trainset.compile(train=True, logger=self)
        if isinstance(testset, Dry.Dataset):
            testset = testset.compile(train=False, logger=self)
        if isinstance(valset, Dry.Dataset):
            valset = valset.compile(train=False, logger=self)
        self.log(f'compiling dataloaders ..')
        self.trainloader = DataLoader(trainset, batch_size=core.batch_size, shuffle=True, num_workers=core.num_workers)
        self.testloader = DataLoader(testset, batch_size=core.batch_size, shuffle=False, num_workers=core.num_workers)
        self.valloader = DataLoader(valset, batch_size=core.batch_size, shuffle=True, num_workers=core.num_workers)
        self.log('compile finished.')

    def epoch_finished(self): return self.status.epoch_finished()
    @property
    def hash(self): return self.status.hash
    @property
    def iter(self): return self.status.iter
    @property
    def epoch(self): return self.status.epoch

    def __step__(self, sample, train=True):
        """
            Args:
                sample: batched sample from dataloader.
                train: train=True if step for train.
        """
        x, y = self.read_batch(sample)
        if train:
            self.model.train()
            self.optimizer.zero_grad()
            logits = self.forward(x)
            loss = self.loss_function(logits, y)
            loss.backward()
            self.optimizer.step()
        else:
            self.model.eval()
            with torch.no_grad():
                logits = self.forward(x)
                loss = self.loss_function(logits, y)
        return loss.item()

    def __train__(self, start_epoch=0):
        """
            train over epoches infinitely (not really),
            Args:
                start_epoch: epoch to start from.
        """
        epoch = start_epoch-1
        while True:
            epoch += 1
            last_iter = 0
            loss = 0.0
            for i, sample in enumerate(self.trainloader, 0):
                loss += self.__step__(sample, train=True)
                epoch_finished = False
                if (i+1) == len(self.trainloader):
                    epoch_finished = True
                if epoch_finished or i-last_iter >= self.step_iters:
                    loss /= i-last_iter
                    yield dict(loss=loss, epoch=epoch, iter=i+1, epoch_finished=epoch_finished)
                    last_iter = i
                    loss = 0.0

    def __val__(self):
        """
            evaluate eval_loss for fixed batches.
        """
        last_iter = 0
        current_iter = 0
        while True:
            loss = 0.0
            for sample in self.valloader:
                loss += self.__step__(sample, train=False)
                current_iter += 1
                if current_iter-last_iter >= self.val_iters:
                    loss /= current_iter-last_iter
                    yield dict(loss=loss)
                    last_iter = current_iter
                    loss = 0.0

    def init_seed(self, seed):
        self.log(f'init with seed = {seed}')
        manual_seed(seed)

    def log(self, line, time_head=True):
        if not self.core.auto_log:
            return
        if self.under_test:
            line = f'testing# {line}'
        if time_head:
            line = f'[{time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())}] {line}'
        print(line)
        fp = self.core.slave_log_path
        if fp is None:
            fp = f'{self.root}/slave.log'
        with open(fp, 'a') as f:
            f.write(f'{line}\n')

    def netdata_log(self, metrics_dict):
        if not self.core.auto_log:
            return
        if self.core.netdata_log_path is not None:
            self.netdata_state.update(metrics_dict)
            with open(self.core.netdata_log_path, "w")as f:
                json.dump(self.netdata_state, f)

    def hyper_log(self, msg, time_head=True, hash_head=True, iter_head=True):
        if not self.core.auto_log:
            return
        if type(msg) is dict:
            for k, v in msg.items():
                self.hyper_log(f'{k}: {v}', time_head=time_head, hash_head=hash_head, iter_head=iter_head)
            return self
        if iter_head:
            msg = f'[{self.status.epoch}, {self.status.iter:5d}/{len(self.trainloader)}] {msg}'
        if hash_head:
            msg = f'<{self.status.hash}>\t{msg}'
        self.log(msg, time_head=time_head)

    def __setinterval__(self, iters=None, val_iters=None):
        if iters:
            assert iters > 0
            if iters < 1:
                iters = int(len(self.trainloader)*iters)
            self.step_iters = iters
        if val_iters:
            assert val_iters > 0
            if val_iters < 1:
                val_iters = int(len(self.valloader)*iters)
            self.val_iters = val_iters

    def steps(self, iters=None, val_iters=None, max_epoch=None):
        '''
            Args:
                iters: set a checkpoint per <iter> iterations.
                val_iters: total validation batches for each checkpoint.
                max_epoch:
        '''
        self.__setinterval__(iters=iters, val_iters=val_iters)
        while True:
            yield self.step()
            if max_epoch is not None and self.status.epoch > max_epoch:
                break

    def step(self, iters=None, val_iters=None):
        """
            step to the next checkpoint, works as a generator.
            yields: status
            Args:
                iters:
                val_iters:
        """
        self.__setinterval__(iters=iters, val_iters=val_iters)
        if self.__train_generator is None:
            self.__train_generator = self.__train__(start_epoch=self.status.epoch)
            self.__val_generator = self.__val__()

        state_train = next(self.__train_generator)
        state_val = next(self.__val_generator)
        self.status.train.update(state_train)
        self.status.val.update(state_val)
        self.hyper_log(f'train_loss: {state_train["loss"]}\tval_loss: {state_val["loss"]}')
        self.netdata_log(dict(train_loss=state_train["loss"], val_loss=state_val["loss"]))
        return self.status

    def check(self, iters=1):
        self.under_test = True
        self.hyper_log(f':: starting check ..', iter_head=False)
        self.step(iters=iters, val_iters=1)
        self.__train_generator = None
        self.__val_generator = None
        self.hyper_log(f':: check passed.', iter_head=False)
        self.under_test = False

    def eval(self, _type=None):
        self.model.eval()
        if _type == 'acc@top1':
            self._eval_hooks = [('acc@top1', functional.accuracy)]
        if _type == 'acc@top3':
            self._eval_hooks = [('acc@top3', functional.accuracy_top3)]
        metrics = {k[0]: 0 for k in self._eval_hooks}
        with torch.no_grad():
            for sample in self.testloader:
                x, y = self.read_batch(sample)
                logits = self.forward(x)
                for name, hook_fn in self._eval_hooks:
                    metrics[name] += hook_fn(logits, y)/len(self.testloader)
        self.hyper_log(metrics)
        self.netdata_log(metrics)
        if _type is not None:
            return metrics[_type]
        return metrics

    def metrics_hook(self, name, hook_fn):
        '''
            Args:
                hook_fn: hook_fn(logits, y) -> metrics_value
        '''
        self._eval_hooks.append((name, hook_fn))

    def forward(self, x):
        """
            Args:
                x:
            override if needed.
        """
        return self.model(x)

    @staticmethod
    def loss_function(logits, y):
        """
            Args:
                logits:
                y:

            override if needed.
        """
        return torch.nn.functional.cross_entropy(logits, y)

    @staticmethod
    def read_batch(sample):
        """
            read x,y from sample,
            override if needed.
        """
        x, y = sample
        return Variable(x).cuda(), Variable(y).cuda()



import inspect


class MissingArgsException(Exception):
    def __init__(self, key, msg):
        super().__init__()
        msg = f'Missing key \'{key}\' in config\n({msg})'

    def __str__(self):
        return f'{self.msg}'


class Dry():
    class Args():
        def __init__(self, compile_class):
            parameters = inspect.signature(compile_class).parameters
            parameter_dict = {k: self.__defaultargs__(v) for k, v in parameters.items()}
            self.compile_class = compile_class
            self.__dict__.update(parameter_dict)

        def update(self, **kwargs):
            self.__dict__.update(kwargs)

        @staticmethod
        def __defaultargs__(v):
            if v.default is not inspect._empty:
                return v.default
            if v.kind == inspect.Parameter.VAR_KEYWORD:
                return {}
            if v.kind == inspect.Parameter.VAR_POSITIONAL:
                return []

        def __str__(self):
            return str(self.__dict__)

        def log(self, msg, logger=None):
            if logger is None:
                print(msg)
            else:
                logger(msg)

        def compile(self, logger=None, **kwargs):
            args = {k: v for k, v in self.__dict__.items() if k != 'compile_class' and k != 'kwargs'}
            args.update(kwargs)
            try:
                self.log(f'==> Compiling: {self.compile_class}', logger=logger)
                for k, v in args.items():
                    if isinstance(v, Dry.Args):
                        args[k] = v = v.compile(logger=logger)
                    self.log(f'  -> Running args hook: {k} = {v}', logger=logger)
                    assert v is not inspect._empty
            except Exception:
                raise MissingArgsException(k, 'This value must be specified.')
            return self.compile_class(**args)

    class Optimizer(Args):
        def __init__(self, compile_class):
            super().__init__(compile_class)
            self.lr = 1e-3
            self.weight_decay = 5e-4

    class Dataset(Args):
        def __init__(self, compile_class):
            super().__init__(compile_class)
            self.root = None
            self.transform = None




import matplotlib
import matplotlib.pyplot as plt
import cv2
import numpy as np
import torch


def numpy_image(x, imtype=np.uint8, mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225)):
    """"将tensor的数据类型转成numpy类型，并反归一化.

    Parameters:
        input_image (tensor) --  输入的图像tensor数组
        imtype (type)        --  转换后的numpy的数据类型
    """

    if not isinstance(x, np.ndarray):
        if isinstance(x, torch.Tensor):  # get the data from a variable
            if len(x.shape) == 4:
                x = x[0]
            image_tensor = x.data
        else:
            return x
        image_numpy = image_tensor.cpu().float().numpy()  # convert it into a numpy array
        if image_numpy.shape[0] == 1:  # grayscale to RGB
            image_numpy = np.tile(image_numpy, (3, 1, 1))
        for i in range(len(mean)):
            image_numpy[i] = image_numpy[i] * std[i] + mean[i]
        image_numpy = image_numpy * 255
        image_numpy = np.transpose(image_numpy, (1, 2, 0))  # post-processing: tranpose and scaling
    else:  # if it is a numpy array, do nothing
        image_numpy = x
    return image_numpy.astype(imtype)


def heatmap(x):
    if len(x.shape) == 4:
        x = x[0]
    assert x.shape[0] == 1
    heatmap = x[0].cpu().numpy()
    heatmap = heatmap/np.max(heatmap)
    # must convert to type unit8
    heatmap = np.uint8(255 * heatmap)
    heatmap = cv2.applyColorMap(heatmap, cv2.COLORMAP_JET)
    return heatmap


def save(x, path):
    assert isinstance(x, np.ndarray)
    matplotlib.image.imsave(path, x)
    # fig = plt.gcf()  # generate outputs
    # plt.imshow(x, aspect='equal'), plt.axis('off'), fig.set_size_inches(size[1]*12/300.0, size[0]*12/300.0)
    # plt.gca().xaxis.set_major_locator(plt.NullLocator()), plt.gca().yaxis.set_major_locator(plt.NullLocator()), plt.subplots_adjust(top=1, bottom=0, left=0, right=1, hspace=0, wspace=0), plt.margins(0, 0)
    # plt.text(0, 0, annotation, color='white', size=4, ha="left", va="top", bbox=dict(boxstyle="square", ec='black', fc='black'))
    # plt.savefig(path, dpi=300, pad_inches=0)    # visualize masked image
    # plt.close('all')



from .accuracy import accuracy
from .smooth import smooth
from .step import train_step, validation_step
from .seed import manual_seed

__all__ = ['accuracy', 'smooth', 'train_step', 'validation_step', 'manual_seed']



import shutil
import tarfile
import os

import re


def collect_hash(lines):
    _cache = []
    for line in lines:
        res = re.search(r'^(?P<value>\w(\w|\d){7})(\.|\-)', line)
        if res is None:
            continue
        res = res.groupdict()['value']
        if res in _cache:
            continue
        _cache.append(res)
        yield res


def clean(source_dir='build', target_dir='dist'):
    check_dir(target_dir)
    source_files = [x for x in os.listdir(source_dir)]
    for x in collect_hash(source_files):
        tar_path = f'{target_dir}/{x}.tar.gz'
        print(f'compress: {tar_path}')
        with tarfile.open(tar_path, "w:gz") as tar:
            for root, _, files in os.walk(source_dir):
                for fname in files:
                    if x in fname:
                        pathfile = os.path.join(root, fname)
                        tar.add(pathfile)
        for fname in source_files:
            if x in fname:
                rm_path = f'{source_dir}/{fname}'
                print(f'remove: {rm_path}')
                if os.path.isdir(rm_path):
                    shutil.rmtree(rm_path, True)
                else:
                    os.remove(rm_path)
    print('clean complete.')


def check_dir(path):
    if not os.path.exists(path):
        os.makedirs(path)
        print(f'created dir: {path}')


def pack(source_dir, cache_dir, name='deploy_pack'):
    check_dir(cache_dir)
    target_fp = f'{cache_dir}/{name}.tar.gz'
    total = 0
    for _, _, files in os.walk(source_dir):  # 遍历统计
        for x in files:
            total += 1  # 统计文件夹下文件个数
    with tarfile.open(target_fp, "w:gz") as tar:
        current = 0
        for root, _, files in os.walk(source_dir):
            for fname in files:
                current += 1
                pathfile = f'{root}/{fname}'
                arcpath = f'{root.split(source_dir)[-1]}/{fname}'
                tar.add(pathfile, arcname=arcpath)
                print(f'compressing ({current}/{total}): {arcpath}')
    print(f'packing {target_fp}')
    return target_fp



from . import functional


class Accuracy():
    def __init__(self, testloader, model, topk=1, cuda=True, check_type_tensor=True):
        self.testloader = testloader
        self.model = model
        self.topk = topk
        self.cuda = cuda
        self.check_type_tensor = check_type_tensor
        self.history = []
        self._hooked_looper = None

    def __call__(self, info=None):
        acc = functional.accuracy(
            testloader=self.testloader,
            model=self.model,
            topk=self.topk,
            cuda=self.cuda,
            check_type_tensor=self.check_type_tensor,
        )
        if info is not None:
            acc = (info, acc)
        elif self._hooked_looper is not None:
            acc = (self._hooked_looper.axis(), acc)
        self.history.append(acc)
        return acc
